GObject {
	var <>ptr; // Do not change position
	var <callbackResults; // Do not change position

	var <gobjectCallbacks, gobjectCallbackId;

	classvar <wrappers;
	classvar <>fatalWarnings;
	classvar <finalizeActions;

	// ------------------------

	*initClass {
		fatalWarnings = true;
		UI.registerForShutdown({
			// We should (in this order):
			//	1) unwrap all objects 
			//		a) free all pieces of non-RT memory
			//		b) disconnect all sclang callbacks
			//		c) call finalize on all objects
			//	2) destroy all toplevels
			//
			// We can do it in sync because we are going
			// to reboot the interpreter or exit the host application.
			nil
		});
		finalizeActions = IdentityDictionary[];
	}

	*newWithPtr {
		arg inPtr;
		^super.new.gobjectInit(nil).wrap(inPtr)
	}
	*wrap {
		arg inPtr, type, wrapperClass;
		var wrapper;

		if (inPtr.isNil, {
			^nil
		});

		if (inPtr.isInteger.not, {
			Post << "WARNING:\n";
			Post << thisMethod.asString << ": ptr is " << inPtr << "\n";
			Post << "This may mean STACK CORRUPTION\n";
			if (fatalWarnings, { this.verboseHalt });
		});

		if (inPtr < 1000, {
			error("Wrap request for pointer suspiciously small:");
			inPtr.postln;
			this.verboseHalt;
		});

		if (wrappers.isNil, {
			wrappers = Dictionary.new;			
		});

		wrapper = wrappers.at(inPtr);
		if (wrapper.notNil, { ^wrapper });

		if (type.isNil, {
			type = this.getObjectTypeName(inPtr);
		});
		if (type.isNil, { ^nil });
		
		if (wrapperClass.isNil, {
			wrapperClass = type.asSymbol.asClass;
		});

		//["Meta_GObject-wrap: WRAPPER CLASS: ", wrapperClass].postln;

		wrapperClass = wrapperClass ? this;

		^wrapperClass.newWithPtr(inPtr);
	}

	*registerWrapper {
		arg wrapper;
		
		if (wrappers.isNil, {
			wrappers = Dictionary.new;			
		});

		if (wrappers.at(wrapper.ptr).notNil, {
			Post << "*** Attempt to create a second wrapper for the same object ***" << "\n";
			[wrapper.hash, wrappers.at(wrapper.ptr).hash].postln;
			^wrapper
		});

		wrappers.put(wrapper.ptr, wrapper);
		if (wrapper.isKindOf(GtkObject), {
			wrapper.prGtkObjectConnectStandardDestroySignalHandler;
		}, {
			wrapper.prGObjectAddStandardWeakReference;
		});
		^nil
	}
	prGtkObjectConnectStandardDestroySignalHandler {
		_prGtkObjectConnectStandardDestroySignalHandler;
		^this.primitiveFailed
	}
	prGObjectAddStandardWeakReference {
		_prGObjectAddStandardWeakReference;
		^this.primitiveFailed
	}

	*new {
		arg inPtr;
		^super.new.gobjectInit(inPtr);
	}
	*newEmpty {
		^super.new
	}

	gobjectInit {
		arg inPtr;

		ptr = inPtr;
		gobjectCallbacks = Dictionary.new;
		gobjectCallbackId = Routine({ var i = 0; loop({ i = i + 1; i.yield;})});

		callbackResults = Array.new;
	}

	wrap {
		arg inPtr;

		if (inPtr.isNil, {
			error("***** Attempt to set the pointer of a wrapper to nil\n");
			this.verboseHalt
		});

		if (ptr.notNil, {
			error("***** Attempt to re-set the pointer of a wrapper\n");
			this.verboseHalt
		});

		ptr = inPtr;

		GObject.registerWrapper(this)
	}

	dump {
		_prGObjectDump;
 		^this.primitiveFailed
	}
	getTypeName {
		_prGObjectGetTypeName;
 		^this.primitiveFailed
	}
	getClassName {
		_prGObjectGetClassName;
 		^this.primitiveFailed
	}
	*getObjectTypeName {
		arg ptr;
		_prGObjectGetObjectTypeName;
		^this.primitiveFailed
	}

	ref {
		this.ref_request;
		Async.wait
	}
	ref_request {
		_prGObjectRef;
		^this.primitiveFailed
	}

	unref {
		this.unref_request;
		Async.wait
	}
	unref_request {
		_prGObjectUnref;
		^this.primitiveFailed
	}

	unrefAll {
		arg objects;
		objects.do({
			arg object;
			if (object.notNil, {
				object.unref;
			});
		});

	}

	forget {
		GObject.unregister(ptr, this);
		// Note:
		// We do not call finalize here (compare with the destroy handler)
		ptr = 0;
	}
	*unregister {
		arg inPtr, inWrapper;
		var wrapper;
		wrapper = wrappers.at(inPtr);
		if (wrapper.isNil, {
			^true
		}, {
			wrappers.removeAt(inPtr);
			^false
		});
	}

	connect {
		arg name, function ... args;
		^this.prConnect(name, function, args, false)
	}
	connectAfter {
		arg name, function ... args;
		^this.prConnect(name, function, args, true)
	}
	prConnect {
		arg name, function, args, after;
		var id1, id2;

		id1 = gobjectCallbackId.next;

/*		Post	<< "Connecting signal " << name
			<< " for " << this.class.name << " at " << ptr
			<< " with id1 " << id1
			<< "\n";
*/
		id2 = this.prGObjectConnect(name, id1, after); 
		gobjectCallbacks.put(id1, [id2, function, args]);

		^id1

	}

	prGObjectConnect {
		arg name, id;
		_prGObjectConnect;
 		^this.primitiveFailed
	}

	disconnect {
		arg id1;
		var id2, function, extraArgs;

		#id2, function, extraArgs = gobjectCallbacks.at(id1);
		if (id2.notNil, {
			this.prGObjectDisconnect(id2);
			gobjectCallbacks.removeAt(id1);
		})
	}
	prGObjectDisconnect {
		arg id;
		_prGObjectDisconnect;
 		^this.primitiveFailed
	}

	getData {
		arg key;
		_prGObjectGetData;
		^this.primitiveFailed
	}

	prScheduleCallback {
		arg id1, argTags, args;
		var id2, function, mappedArgs, extraArgs;
		var argument, tag;

		/*Post << "****************** " << thisMethod.asString << nl;
		this.dump;
		id1.dump;
		argTags.dump;
		args.dump;
		Post << "++++++++++++++++++" << nl;*/

		Async.run({
			var value;

			#id2, function, extraArgs = gobjectCallbacks.at(id1);
			if (argTags.isNil || args.isNil, {
				mappedArgs = [];
			}, {
				mappedArgs = args.copy;
				argTags.do({
					arg tag, index;
					/*
					Post << "\tid1:\t" << id1 << "\n";
					Post << "\targTags:\t" << argTags << "\n";
					Post << "\targs:\t" << args << "\n\n";
					*/
					if (tag.notNil, {
						mappedArgs.put(index, GObject.wrap(args.at(index)));
					});
				});
			});
		
			//"STARTING CALLBACK".postln;
			function.valueArray(mappedArgs ++ extraArgs);
			
		}, [], {
			arg result;
			callbackResults = callbackResults.add(result);
			//Post << "  CALLBACK RESULT STACK IS NOW: " << callbackResults.size << "\n";
			GtkServer.sendResultTrigger;
			//"CALLBACK TERMINATED".postln;
		}, {
			GtkServer.sendAbortTrigger;
		}, mode: 'callback');
	}

	prDestroySignalHandler {
		//Post << thisMethod.asString << " on " << ptr << " (it's "
		//       << wrappers.at(ptr) << " with hash " << wrappers[ptr].basicHash << ")\n";

		if (ptr == 0, {
			Post << "*** WARNING:\n";
			Post << "    A GObject wrapper of kind " << this.class.asString << " received a destroy signal\n";
			Post << "    but it's pointer is null.\n";
			Post << "    This mean it received TWO destroy signals\n";
		});

		if (GObject.unregister(ptr, this), {
			Post << "*** WARNING:\n";
			Post << "    A GObject wrapper of kind " << this.class.asString << " received a destroy signal\n";
			Post << "    but it is not registered in GObject.wrappers.\n"
		}, {
			wrappers.removeAt(ptr);
		});

		GtkServer.sendVoidTrigger;

		Async.run({
			this.finalize;
			ptr = nil;
		}, [], {
		}, {
			GtkServer.sendAbortTrigger
		}, mode: 'callback');
	}
	prWeakReferenceNotifyHandler {
		^this.prDestroySignalHandler
	}
	hasPendingSignalAnswers {
		^(callbackResults.size > 0)
	}

	asInteger {
		_prGObjectAsInteger;
		^this.primitiveFailed
	}
	alive {
		^(this.asInteger != 0)
	}

	getProperty {
		arg name;
		this.getProperty_request(name);
		^Async.wait
	}
	getProperty_request {
		arg name;
		_prGObjectGetProperty;
		^this.primitiveFailed
	}
	setProperty {
		arg name, value;
		this.setProperty_request(name, value);
		Async.wait
	}
	setProperty_request {
		arg name, value;
		_prGObjectSetProperty;
		^this.primitiveFailed
	}

	// Default finalization method called before we invalidate this wrapper
	finalize {
		arg function;
		function = GObject.finalizeActions.removeAt(ptr);
		function.value;
	}
	finalizeAction_ {
		arg function;
		GObject.finalizeActions[ptr] = function;
	}
}

GtkServer {
   	*sendResultTrigger {
		_prGtkServerSendResultTrigger;
		this.primitiveFailed
	}
   	*sendVoidTrigger {
		_prGtkServerSendVoidTrigger;
		this.primitiveFailed
	}
   	*sendAbortTrigger {
		_prGtkServerSendAbortTrigger;
		this.primitiveFailed
	}
}
