#include <algorithm>

#include <unistd.h>
#include <fcntl.h>

#include <glib/gtimer.h>
#include <gtk/gtkmain.h>

#include "gtkrequest.h"

int GtkRequestQueue::pipein = 0, GtkRequestQueue::pipeout = 0;
unsigned int GtkRequestQueue::slot_size = 0;
char *GtkRequestQueue::stack = 0;
char *GtkRequestQueue::buffer = 0;
int GtkRequestQueue::sp = -1;
int GtkRequestQueue::rlocator = 0;
int GtkRequestQueue::wlocator = 0;
pthread_mutex_t GtkRequestQueue::mutex = PTHREAD_MUTEX_INITIALIZER;
GTimer *GtkRequestQueue::performTimer = 0;
int GtkRequestQueue::depth = 0;
int GtkRequestQueue::triggerInputId = -1;

#define DUMP_MALLOC_STATS
#undef DUMP_STACK_DEPTH_GREATER_THAN_ONE
#undef DUMP_RESULT_BEFORE_IT_IS_DELIVERED

extern unsigned int gtkasyncrequests_get_requested_slot_size(void);

void GtkRequestQueue::initialize(void) {
  int fds[2];
  pipe(fds);
  pipein = fds[0];
  pipeout = fds[1];

  fcntl(pipein, F_SETFL, O_NONBLOCK);
  fcntl(pipeout, F_SETFL, O_NONBLOCK);

  slot_size = gtkasyncrequests_get_requested_slot_size();

  stack = new char[slot_size * GTK_REQUEST_RECURSION_MAX];
  buffer = new char[slot_size * GTK_REQUEST_BUFFER_SIZE];

#ifdef DUMP_MALLOC_STATS
  std::cout << "GtkRequestQueue statistics" << std::endl;
  std::cout << "  Stack size (slots): " << GTK_REQUEST_RECURSION_MAX << std::endl;
  std::cout << "  Ring buffer size (slots): " << GTK_REQUEST_BUFFER_SIZE << std::endl;
  std::cout << "  Size of request stack: " << slot_size * GTK_REQUEST_RECURSION_MAX << std::endl;
  std::cout << "  Size of request buffer: " << slot_size * GTK_REQUEST_BUFFER_SIZE << std::endl;
#endif

  performTimer = g_timer_new();
}

void GtkRequestQueue::finalize(void) {
  close(pipein);
  close(pipeout);
  delete [] stack;
  delete [] buffer;
  g_timer_destroy(performTimer);
}

bool GtkRequestQueue::bufferRelease(void) {
  if (rlocator == wlocator) {
#ifdef GTK_REQUEST_DEBUG
    std::cout << "GtkRequestQueue: ring buffer underflow" << std::endl;
#endif
    return true;
  }
  if (++rlocator >= GTK_REQUEST_BUFFER_SIZE) rlocator = 0;
  return false;
}

// Requests are constructed on the ring buffer, bit-transferred on the stack
// where they destroyed after their performance
void GtkRequestQueue::transfer(int n, int m) {
  void *src = bufferAt<void>(n);
  // Most requests are much smaller then the slot size, just few bytes
  unsigned int real_size = ((GtkRequest *)src)->size();
  //std::cout << "copying " << real_size << " bytes (out of " << slot_size << " ) from " << bufferAt<void>(n) << " to " << at<void>(m) << std::endl;
  memcpy(at<void>(m), src, real_size);
}
  
void GtkRequestQueue::pushFromBuffer(void) {
  int sp = push();
  if (sp < 0) {
#ifdef GTK_REQUEST_DEBUG
    std::cout << "GtkRequestQueue stack overflow" << std::endl;
#endif
  } else {
    transfer(rlocator, sp);
    bufferRelease();
  }
}

unsigned int GtkRequestQueue::slotSize(void) {
  return slot_size;
}

char GtkRequestQueue::readTrigger(void) {
  char buffer[1];
  if (read(pipein, buffer, 1) != 1)
    return '\0';
  //std::cout << "trigger: " << buffer[0] << std::endl;
  return buffer[0];
}

GtkRequestQueue::PollResult GtkRequestQueue::smartWait(int msec) {
  int remainingTime = 0, timeToSleep = msec;
  GtkRequestQueue::PollResult result;

  do {
    remainingTime = 0;
      
    result = GtkRequestQueue::poll(timeToSleep, remainingTime); 

    if (result == GtkRequestQueue::INCOMING_REQUEST) {
#ifdef GTK_REQUEST_DEBUG
      std::cout << "Handling a trigger in SmartWait mode" << std::endl;
#endif
      GtkRequestQueue::handleRequestTrigger();
    } else if (result == GtkRequestQueue::INCOMING_DTOR_REQUEST) {
#ifdef GTK_REQUEST_DEBUG
      std::cout << "Handling a DTOR trigger in SmartWait mode" << std::endl;
#endif
      GtkRequestQueue::handleRequestTrigger(true);
    } else if (result == GtkRequestQueue::TIMEOUT) {
      return result;
    } else if (result == GtkRequestQueue::INCOMING_RESULT) {
      return result;
    } else if (result == GtkRequestQueue::INCOMING_VOID) {
      return result;
    } else if (result == GtkRequestQueue::ABORT) {
      return result;
    }
    timeToSleep = msec - remainingTime;
  } while (timeToSleep > 0);
}

GtkRequestQueue::PollResult GtkRequestQueue::poll(int msec, int &remainingTime) {
  pollfd pfd[1];
  timeval tv1, tv2;

  pfd[0].fd = pipein;
  pfd[0].events = POLLIN;
    
  gettimeofday(&tv1, 0);
  int n = ::poll(pfd, 1, msec);
  gettimeofday(&tv2, 0);

  if (n == 0) {
    remainingTime = 0;
    return TIMEOUT;
  } else if (n > 0) {
    char t;
    remainingTime = ((tv2.tv_sec - tv2.tv_sec) * 1000) + ((tv2.tv_usec - tv2.tv_usec) / 1000);
    t = GtkRequestQueue::readTrigger();
    if (t == '\0') {
#ifdef GTKSERVER_MORE_WARNINGS
      std::cout << "no trigger to read!!!!" << std::endl;
#endif
      return ERROR;
    } else if (t == 'r') {
      return INCOMING_REQUEST;
    } else if (t == 'd') {
      return INCOMING_DTOR_REQUEST;
    } else if (t == 'a') {
      return INCOMING_RESULT;
    } else if (t == 'v') {
      return INCOMING_VOID;
    } else if (t == '%') {
      return ABORT;
    } else {
      std::cout << "unknown trigger!!!!" << std::endl;
      return ERROR;
    }
  } else {
    remainingTime = ((tv2.tv_sec - tv2.tv_sec) * 1000) + ((tv2.tv_usec - tv2.tv_usec) / 1000);
    return ERROR;
  }
}

void GtkRequestQueue::lock(void) {
  pthread_mutex_lock(&mutex);
}
void GtkRequestQueue::unlock(void) {
  pthread_mutex_unlock(&mutex);
}

int GtkRequestQueue::push(void) {
  if (sp < GTK_REQUEST_RECURSION_MAX - 1) {
    return ++sp;
  } else {
    return -1;
  }
}

bool GtkRequestQueue::pop(void) {
  if (sp >= 0) {
    sp -= 1;
    return false;
  } else {
    return true;
  }
}

void GtkRequestQueue::trigger(char t) {
  char buffer[1];

  buffer[0] = t;

  //std::cout << buffer[0] << std::endl;

  if (write(pipeout, buffer, 1) == EAGAIN) {
    std::cout << "triggering again (pathologic)" << std::endl;
  }
#ifdef GTK_REQUEST_DEBUG
  std::cout << "trigger sent" << std::endl;
#endif
}

void GtkRequestQueue::performTop(bool noanswer) {
  GtkRequest *request = GtkRequestQueue::top<GtkRequest>();

#ifdef DUMP_STACK_DEPTH_GREATER_THAN_ONE
  if (sp > 0)
    std::cout << "Async request stack depth: " << sp + 1 << std::endl;
#endif

  if (request == 0) {
    std::cout << "ERROR: GtkRequestQueue attempt to perform under the bottom of the stack" << std::endl;
  } else {
#ifdef GTK_REQUEST_DEBUG
    std::cout << "performing " << ((void *)request) << std::endl;
#endif
    request->perform();
      
    if (!noanswer) {
      int nargs = 1;

      VMGlobals *g = gMainVMGlobals;

      pthread_mutex_lock(&gLangMutex);
      bool canCallOS = g->canCallOS;
      g->canCallOS = true;
      PyrClass *object = getsym("Async")->u.classobj;
      ++g->sp; SetObject(g->sp, object);

      nargs += request->pushResult(g);

#ifdef DUMP_RESULT_BEFORE_IT_IS_DELIVERED
      {
	char buffer[1024];
	request->code_location(buffer);

	if (nargs > 1) {
	  int i, k = nargs - 1;
	  post("AsyncRequest at %s is returning %d object(s):\n", buffer, k);
	  for (i = 0; i < k; ++i) {
	    post("%d of %d:\n", i, k);
	    dumpObjectSlot(g->sp - (k - 1 + i));
	  }
	} else {
	  post("AsyncRequest at %s isn't returning anything\n", buffer);
	}
      }
#endif
      runInterpreter(g, getsym("deliverRequestResult"), nargs);

      g->canCallOS = canCallOS;
      pthread_mutex_unlock(&gLangMutex);
    }

    if (GtkRequestQueue::pop()) {
      std::cout << "ERROR: GtkRequestQueue attempt to perform under the bottom of the stack (2)" << std::endl;
    } else {
      request->~GtkRequest();
    }
  }
}

void GtkRequestQueue::handleRequestTrigger(bool noanswer) {
#ifdef GTK_REQUEST_DEBUG
  std::cout << "handling a request trigger" << std::endl;
#endif
  GtkRequestQueue::lock();
  pushFromBuffer();
  GtkRequestQueue::unlock();
  GtkRequestQueue::performTop(noanswer);
#ifdef GTK_REQUEST_PROFILE
  gulong usec;
  std::cout << "DT: " << g_timer_elapsed(performTimer, &usec) << std::endl;
  g_timer_reset(performTimer);
#endif
#ifdef GTK_REQUEST_DEBUG
  std::cout << "complete trigger handling!" << std::endl;
#endif
}

bool GtkRequestQueue::triggerHandler(GdkInputCondition condition) {
  if (condition == GDK_INPUT_READ) {
    char t = GtkRequestQueue::readTrigger();
    if (t == '\0') {
#ifdef GTKSERVER_MORE_WARNINGS
      std::cout << "no trigger to read!!!!" << std::endl;
#endif
      return ERROR;
    } else if (t == 'r') {
      handleRequestTrigger();
    } else if (t == 'd') {
      handleRequestTrigger(true);
    } else if (t == 'a') {
      std::cout << "GtkServerERROR: callback answer received in main loop!" << std::endl;
    } else if (t == 'v') {
      std::cout << "GtkServerERROR: VOID answer received in main loop!" << std::endl;
    } else if (t == '%') {
      std::cout << "GtkServerERROR: abort received in main loop!" << std::endl;
    } else {
      std::cout << "unknown trigger!!!!" << std::endl;
    }
  } else {
    std::cout << "exceptional I/O condition on GtkRequestQueue pipe" << std::endl;
  }
  return true;
}

void GtkRequestQueue::disconnectMainloopCallback(void) {
  if (triggerInputId > 0) gtk_input_remove(triggerInputId);
}

static void call_trigger_handler(gpointer data, gint source, GdkInputCondition condition) {
  GtkRequestQueue::triggerHandler(condition);
}

void GtkRequestQueue::connectMainloopCallback(void) {
  triggerInputId = gtk_input_add_full(pipein, (GdkInputCondition)(GDK_INPUT_READ | GDK_INPUT_EXCEPTION), call_trigger_handler, 0, 0, 0);
}
