#ifndef _asyncrequestqueue_H_
#define _asyncrequestqueue_H_

#include <iostream>
#include <cstdlib>

#include <poll.h>
#include <unistd.h>
#include <sys/time.h>

#include <sigc++/sigc++.h>

#include <glib/giochannel.h>
#include <glib/gtimer.h>

#include <SCBase.h>
#include <PyrSlot.h>
#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <PyrMessage.h>
#include <GC.h>
#include <VMGlobals.h>
extern pthread_mutex_t gLangMutex;

#define LARGE_STRING_SIZE 10240
#define MEDIUM_STRING_SIZE 1024
#define SMALL_STRING_SIZE 128
#define FILENAME_STRING_SIZE 1024

#undef GTK_REQUEST_PROFILE
#undef GTK_REQUEST_DEBUG
#undef GTKSERVER_MORE_WARNINGS
#undef GTKSERVER_FATAL_FAILURES

inline int safe_string_copy(char *dest, const char *source, int size, int space) {
  if (size < 0) size = strlen(source);
  strncpy(dest, source, space);
  if (size >= space) size = space - 1;
  dest[size] = '\0';
  return size;
}

class GtkRequest {
public:
  GtkRequest(void) { }
  virtual ~GtkRequest() { }

  virtual void perform(void) { };
  virtual int pushResult(VMGlobals *g) {
    return 0;
  }
  virtual unsigned int size(void) const = 0;
  virtual const char *name(void) const { return "unknown"; };
  virtual void code_location(char *buffer) const { strcpy(buffer, "unknown"); };
};

#define SIZE_METHOD_DEFINITION unsigned int size(void) const { return sizeof(*this); }
#define NAME_METHOD_DEFINITION(reqname) const char *name(void) const { return #reqname ; }

#define CODE_LOCATION_METHOD_DEFINITION void code_location(char *buffer) const { sprintf(buffer, "%s:%d", __FILE__, __LINE__); }

class GtkRequest_ReturnBoolean : public GtkRequest {
public:
  int pushResult(VMGlobals *g) {
    ++g->sp;
    if (m_value) {
      SetTrue(g->sp);
    } else {
      SetFalse(g->sp);
    }
    return 1;
  }

protected:
  bool m_value;
};

class GtkRequest_ReturnInteger : public GtkRequest {
public:
  int pushResult(VMGlobals *g) {
    ++g->sp;
    SetInt(g->sp, m_value);
    return 1;
  }

protected:
  int m_value;
};

// ---------------------------------

#define GTK_REQUEST_RECURSION_MAX 64
#define GTK_REQUEST_BUFFER_SIZE 64
#define GTK_REQUEST_SLOT_AT(n, TYPE) ((TYPE *)((stack + slot_size * n)))
#define GTK_REQUEST_BUFFER_SLOT_AT(n, TYPE) ((TYPE *)((buffer + slot_size * n)))

class GtkRequestQueue {
public:
  static void initialize(void);
  static void finalize(void);

  template <class ReqType> static ReqType *at(int n) {
    if (sp >= 0 && n < GTK_REQUEST_RECURSION_MAX)
      return GTK_REQUEST_SLOT_AT(n, ReqType);
    else
      return 0;
  }

  template <class ReqType> static ReqType *bufferAt(int n) {
    if (n >= 0 && n < GTK_REQUEST_BUFFER_SIZE)
      return GTK_REQUEST_BUFFER_SLOT_AT(n, ReqType);
    else
      return 0;
  }

  template <class ReqType> static ReqType *top(void) {
    if (sp < 0)
      return 0;
    else
      return GTK_REQUEST_SLOT_AT(sp, ReqType);
  }

  template <class ReqType> static ReqType *bufferAllocate(void) {
    if (wlocator == (rlocator - 1) || (wlocator == (GTK_REQUEST_BUFFER_SIZE - 1) && rlocator == 0)) {
#ifdef GTK_REQUEST_DEBUG
      std::cout << "GtkRequestQueue: ring buffer overflow" << std::endl;
#endif
      return 0;
    }

    int n = wlocator;
    if (++wlocator >= GTK_REQUEST_BUFFER_SIZE) wlocator = 0;
    
    //std::cout << "allocating slot " << n << " from buffer (" << bufferAt<void>(n) << ")" << std::endl;

    return bufferAt<ReqType>(n);
  }

  static bool bufferRelease(void);
  static void transfer(int n, int m); 
  static void pushFromBuffer(void);
  static unsigned int slotSize(void);

  enum PollResult {
    TIMEOUT,
    ERROR,
    INCOMING_REQUEST,
    INCOMING_DTOR_REQUEST,
    INCOMING_RESULT,
    INCOMING_VOID,
    ABORT
  };

  static char readTrigger(void);
  static GtkRequestQueue::PollResult smartWait(int msec);
  static PollResult poll(int msec, int &remainingTime);
  static void lock(void);
  static void unlock(void);
  static int push(void);
  static bool pop(void);

  static void trigger(char t = 'r');
  static void performTop(bool noanswer = false);
  static void handleRequestTrigger(bool noanswer = false);
  static bool triggerHandler(GdkInputCondition condition);
  static void disconnectMainloopCallback(void);
  static void connectMainloopCallback(void);

private:
  static int pipein, pipeout;
  static unsigned int slot_size;
  static char *stack;
  static char *buffer;
  static int sp, rlocator, wlocator;
  static pthread_mutex_t mutex;
  static int triggerInputId;
  static GTimer *performTimer;
  static int depth;
};

#define GTK_REQUEST_QUEUE_SEND_OR_FAIL(request_ctor) {   \
    void *slot;                                          \
    GtkRequestQueue::lock();                             \
    slot = GtkRequestQueue::bufferAllocate<void>();      \
    if (slot != 0) {                                     \
      new(slot) request_ctor;                            \
      GtkRequestQueue::trigger();                        \
    }                                                    \
    GtkRequestQueue::unlock();                           \
    if (slot == 0) {                                     \
      error("********* GtkRequestQueue ring buffer overflow **********+\n");   \
    }                                                    \
  }

#define GTK_REQUEST_QUEUE_INIT_SEND_OR_FAIL(request_ctor, i_n_i_t__c_o_d_e) {   \
    void *slot;                                                                 \
    GtkRequestQueue::lock();                                                    \
    slot = GtkRequestQueue::bufferAllocate<void>();                             \
    if (slot != 0) {                                                            \
      new(slot) request_ctor;                                                   \
      i_n_i_t__c_o_d_e;                                                         \
      GtkRequestQueue::trigger();                                               \
    }                                                                           \
    GtkRequestQueue::unlock();                                                  \
    if (slot == 0) {                                                            \
      error("GtkRequestQueue ring buffer overflow\n");                          \
    }                                                                           \
  }

#define GTK_REQUEST_QUEUE_SEND_OR_FAIL_WITHOUT_ANSWER(request_ctor) {   \
    void *slot;                                          \
    GtkRequestQueue::lock();                             \
    slot = GtkRequestQueue::bufferAllocate<void>();      \
    if (slot != 0) {                                     \
      new(slot) request_ctor;                            \
      GtkRequestQueue::trigger('d');                        \
    }                                                    \
    GtkRequestQueue::unlock();                           \
    if (slot == 0) {                                     \
      error("********* GtkRequestQueue ring buffer overflow **********+\n");   \
    }                                                    \
  }

#endif // _asyncrequestqueue_H_
