#include <iostream>
#include <memory>

#include <gtk/gtkmain.h>

#include "sclangInterpreter.h"
#include "simpleConfigFile.h"

#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <GC.h>
#include <VMGlobals.h>

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "gtkbindings.h"
#include "xml2bindings.h"
#include "asyncPrimitives.h"

#undef POST_TO_STDOUT

// XXX: we should use std::auto_ptr for the requests
//      but GCC can't handle local types as template arguments

void setPostFile(FILE* file);
extern "C" int vpost(const char *fmt, va_list vargs);
void post(const char *fmt, ...);
void postfl(const char *fmt, ...);
void postText(const char *text, long length);
void postChar(char c);
void error(const char *fmt, ...);
void flushPostBuf();

void closeAllGUIScreens();
void initGUI();
void initGUIPrimitives();
void initSCViewPrimitives();
void initCocoaFilePrimitives();
void initRendezvousPrimitives();
void initSpeechPrimitives();

long scMIDIout(int port, int len, int statushi, int chan, int data1, int data2);

extern PyrString* newPyrStringN(class GC *gc, long length, long flags, long collect);

extern long compiledOK;
extern pthread_mutex_t gLangMutex;
extern SC_LibraryConfig *gLibraryConfig;

namespace {
  const int SC_VMDefaultMemSpace = 2*1024*1024;
  const int SC_VMDefaultMemGrow	= 256*1024;
  const int SC_VMDefaultNetworkPort = 57120;

  PyrSymbol *s_tick = 0;

  SclangInterpreter *SclangInterpreter::m_self = 0;

}

void SclangInterpreter::vm_start(void) {
  pyr_init_mem_pools(SC_VMDefaultMemSpace, SC_VMDefaultMemGrow);
  init_OSC(57120);
  schedInit();
  gLibraryConfig = SimpleConfigFile::loadFromStandardPaths();
  compileLibrary();
}

void SclangInterpreter::vm_stop(void) {
  schedStop();
}

#define DIVIDER_FACTOR 1

int SclangInterpreter::vm_tick(void) {
  int res = -1;
  static int divider = 0;

  if (++ divider < DIVIDER_FACTOR) return -2;

  divider = 0;

  if (pthread_mutex_trylock(&gLangMutex) == 0) {
    if (compiledOK) {
      runLibrary(s_tick);
      res = 0;
    }
    pthread_mutex_unlock(&gLangMutex);
  }

  flushPostBuf();

  return res;
}

int SclangInterpreter::eval_string_sync(const std::string &s) {
  int res = -1;

  if (compiledOK) {
    pthread_mutex_lock(&gLangMutex);
    VMGlobals *g = gMainVMGlobals;
    
    PyrString* strobj = newPyrStringN(g->gc, s.size(), 0, true);
    memcpy(strobj->s, s.c_str(), s.size());
    
    ++g->sp; SetObject(g->sp, getsym("Host")->u.classobj);
    ++g->sp; SetObject(g->sp, strobj);
    runInterpreter(g, getsym("interpretPrintCmdLineAsyncronously"), 2);

    pthread_mutex_unlock(&gLangMutex);
  }

  return res;
}

int SclangInterpreter::eval_string_async(const std::string &s) {
  int res = -1;

  if (compiledOK) {
    pthread_mutex_lock(&gLangMutex);
    VMGlobals *g = gMainVMGlobals;
    
    PyrString* strobj = newPyrStringN(g->gc, s.size(), 0, true);
    memcpy(strobj->s, s.c_str(), s.size());
    
    SetObject(&g->process->interpreter.uoi->cmdLine, strobj);
    g->gc->GCWrite(g->process->interpreter.uo, strobj);
    
    runLibrary(getsym("interpretPrintCmdLine"));

    pthread_mutex_unlock(&gLangMutex);
  }

  return res;
}

SclangInterpreter::SclangInterpreter(void)
{
  if (m_self != 0) {
    std::cout << "SclangInterpreter is a singleton" << std::endl;
    exit(1);
  }

  m_self = this;

  m_consoleBuffer = gtk_text_buffer_new(0);
  gtk_text_buffer_create_tag(m_consoleBuffer, "command", "weight", PANGO_WEIGHT_BOLD, 0);
  gtk_text_buffer_create_tag(m_consoleBuffer, "result", 0);
  gtk_text_buffer_create_tag(m_consoleBuffer, "result-bullet", "foreground", "red", 0);
}

SclangInterpreter::~SclangInterpreter() {
  std::list<std::string>::iterator it = m_tempFiles.begin();
  while (it != m_tempFiles.end()) unlink((*it++).c_str());
  m_self = 0;
  g_object_unref((gpointer)m_consoleBuffer);
}

void SclangInterpreter::post(const std::string &text) {
  struct SclangConsoleOutput : public Callable {
    std::string m_text;

    SclangConsoleOutput(const std::string &text) : m_text(text) { }
    void operator()(void) {
      GtkTextBuffer *buffer = SclangInterpreter::get_instance()->consoleBuffer();
      if (buffer) {
	GtkTextIter iter;
	gtk_text_buffer_get_end_iter(buffer, &iter);
	gtk_text_buffer_insert(buffer, &iter, m_text.c_str(), m_text.size());
      } else {
	std::cout << m_text << std::flush;
      }
    }
  };

  //std::cout << "posting: ->>" << text << "<<-" << std::endl;

#ifndef POST_TO_STDOUT
  if (m_self)
    m_self->m_callableQueue.post(new SclangConsoleOutput(text));
  else
    std::cout << text << std::flush;
#else
  std::cout << text << std::flush;
#endif
}

void SclangInterpreter::test(void) {
  vm_start();
  sleep(10000);
}

SclangInterpreter *SclangInterpreter::get_instance(void) { return m_self; }

gboolean SclangInterpreter::call_clock_timeout(gpointer data) {
  SclangInterpreter *self = SclangInterpreter::get_instance();
  self->on_clock_timeout();
  return TRUE;
}

int SclangInterpreter::initialize(void) {
  gtk_timeout_add(50, call_clock_timeout, 0);
  vm_start();
  return 0;
}

void SclangInterpreter::reset(void) {
  compileLibrary();
}

void SclangInterpreter::finalize(void) {
  vm_stop();
}

bool SclangInterpreter::on_clock_timeout(void) {
  vm_tick();
  return true;
}

int SclangInterpreter::make_tmp_file(void) {
  char* tmpFile = 0;
  GError* err = 0;
  
  int fd = g_file_open_tmp("SuperCollider.XXXXXX", &tmpFile, &err);
  if (fd >= 0) {
    m_self->m_tempFiles.push_back(std::string(tmpFile));
  }
  
  g_free(tmpFile);
  
  return fd;
}

const char* SclangInterpreter::get_last_tmp_file(void) {
  return m_self->m_tempFiles.empty() ? 0 : m_self->m_tempFiles.back().c_str();
}

int slotStrVal(PyrSlot *slot, std::string &s) {
  if (IsSym(slot)) {
    s = slot->us->name;
    return errNone;
  } else if (isKindOfSlot(slot, class_string)) {
    s = std::string(slot->uos->s, slot->uos->s + slot->uo->size);
    return errNone;
  }
  return errWrongType;
}

int SclangInterpreter::prMetacNullPrimitive(struct VMGlobals *g, int numArgsPushed) {
  return errNone;
}

int slotGetRect(PyrSlot* a, float &x, float &y, float &w, float &h) {
  PyrSlot *slots = a->uo->slots;
  int err;

  err = slotFloatVal(slots+0, &x);
  if (err) return err;
  err = slotFloatVal(slots+1, &y);
  if (err) return err;
  err = slotFloatVal(slots+2, &w);
  if (err) return err;
  err = slotFloatVal(slots+3, &h);
  if (err) return err;
  
  return errNone;
}

int SclangInterpreter::prSCWindow_New(struct VMGlobals *g, int numArgsPushed) {
  return errNone;
}

void SclangInterpreter::initialize_primitives(void) {
  int base = nextPrimitiveIndex();
  int index = 0;

  definePrimitive(base, index++, "_NewTextWindow", prMetacNullPrimitive, 3, 0);
  definePrimitive(base, index++, "_OpenTextFile", prMetacNullPrimitive, 3, 0);

  // Unimplemented primitives:
  definePrimitive(base, index++, "_GetStringFromUser", prMetacNullPrimitive, 3, 0);	
  definePrimitive(base, index++, "_SCWindow_New", prSCWindow_New, 6, 0);	
  definePrimitive(base, index++, "_SCWindow_Refresh", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_Close", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_Minimize", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_ToFront", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_BeginFullScreen", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_EndFullScreen", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_SetShouldClose", prMetacNullPrimitive, 2, 0);	
    
  definePrimitive(base, index++, "_SCWindow_Show", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_Hide", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_SCWindow_SetName", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_SCWindow_SetAlpha", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_SCWindow_SetBounds", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_SCWindow_GetBounds", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_SCWindow_GetScreenBounds", prMetacNullPrimitive, 2, 0);	

  definePrimitive(base, index++, "_String_GetBounds", prMetacNullPrimitive, 3, 0);	
  definePrimitive(base, index++, "_String_DrawAtPoint", prMetacNullPrimitive, 4, 0);	
  definePrimitive(base, index++, "_String_DrawInRect", prMetacNullPrimitive, 4, 0);	
  definePrimitive(base, index++, "_Color_SetStroke", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_Color_SetFill", prMetacNullPrimitive, 1, 0);	

  definePrimitive(base, index++, "_Pen_StrokeRect", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_FillRect", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_StrokeOval", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_FillOval", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_StrokeLine", prMetacNullPrimitive, 3, 0);	
  definePrimitive(base, index++, "_Pen_Push", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_Pen_Pop", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_Pen_Translate", prMetacNullPrimitive, 3, 0);	
  definePrimitive(base, index++, "_Pen_Scale", prMetacNullPrimitive, 3, 0);	
  definePrimitive(base, index++, "_Pen_Rotate", prMetacNullPrimitive, 4, 0);	
  definePrimitive(base, index++, "_Pen_SetWidth", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_SetMatrix", prMetacNullPrimitive, 2, 0);	

  definePrimitive(base, index++, "_Pen_BeginPath", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_Pen_MoveTo", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_LineTo", prMetacNullPrimitive, 2, 0);	
  definePrimitive(base, index++, "_Pen_StrokePath", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_Pen_FillPath", prMetacNullPrimitive, 1, 0);	
        
  definePrimitive(base, index++, "_Pen_DrawAquaButton", prMetacNullPrimitive, 5, 0);	
        
  definePrimitive(base, index++, "_HasDeferred", prMetacNullPrimitive, 1, 0);	
  definePrimitive(base, index++, "_LoadUserPanel", prMetacNullPrimitive, 2, 0);	

  definePrimitive(base, index++, "_Rect_Draw", prMetacNullPrimitive, 3, 0);
  definePrimitive(base, index++, "_Oval_Draw", prMetacNullPrimitive, 3, 0);
  //definePrimitive(base, index++, "_Lines_Draw", prLines_Draw, 1, 0);
  //definePrimitive(base, index++, "_ZigZag_Draw", prZigZag_Draw, 1, 0);
  //definePrimitive(base, index++, "_Polygon_Draw", prPolygon_Draw, 1, 0);
  //definePrimitive(base, index++, "_String_GetBounds", prString_GetBounds, 2, 0);
  //definePrimitive(base, index++, "_String_Draw", prString_Draw, 1, 0);
  //definePrimitive(base, index++, "_Symbol_Draw", prSymbol_Draw, 1, 0);
  //definePrimitive(base, index++, "_Char_Draw", prChar_Draw, 1, 0);
  definePrimitive(base, index++, "_TextBox_Draw", prMetacNullPrimitive, 5, 0);
  definePrimitive(base, index++, "_RoundRect_Draw", prMetacNullPrimitive, 3, 0);
  definePrimitive(base, index++, "_BevelRect_Draw", prMetacNullPrimitive, 3, 0);
  definePrimitive(base, index++, "_Button_Draw", prMetacNullPrimitive, 6, 0);
  definePrimitive(base, index++, "_Arc_Draw", prMetacNullPrimitive, 3, 0);
  definePrimitive(base, index++, "_Font_AvailableFonts", prMetacNullPrimitive, 1, 0);

  initialize_gtk_primitives();
  initialize_xml2_primitives();
  initialize_async_primitives();
}

// =====================================================================
// Library functions

void setPostFile(FILE* file) {
}

int vpost(const char *fmt, va_list vargs) {
  char buffer[1024];
  vsnprintf(buffer, 1024, fmt, vargs);
  SclangInterpreter::post(buffer);
}

void post(const char *fmt, ...) {
  va_list vargs;
  va_start(vargs, fmt); 
  vpost(fmt, vargs);
}

void postfl(const char *fmt, ...) {
  va_list vargs;
  va_start(vargs, fmt); 
  vpost(fmt, vargs);
  flushPostBuf();
}

void postText(const char *text, long len) {
  std::string buffer(text, text + len);
  SclangInterpreter::post(buffer);
}

void postChar(char c) {
  std::string buffer;
  buffer += c;
  SclangInterpreter::post(buffer);
}

void error(const char *fmt, ...) {
  va_list vargs;
  va_start(vargs, fmt);
  SclangInterpreter::post("Error: ");
  vpost(fmt, vargs);
}

void flushPostBuf(void) {
  fflush(stdout);
}

void initGUI(void) {
}

void initGUIPrimitives() {
  s_tick = getsym("tick");
  SclangInterpreter::initialize_primitives();
}

void initSCViewPrimitives() {
}

void initCocoaFilePrimitives() {
}

void initRendezvousPrimitives() {
}

void initSpeechPrimitives() {
}

long scMIDIout(int port, int len, int statushi, int chan, int data1, int data2) {
  return 0;
}
