#include <iostream>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <PyrObject.h>
#include <PyrKernel.h>
#include <PyrPrimitive.h>
#include <GC.h>
#include <VMGlobals.h>

int prXmlDocNew(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *versionSlot = g->sp;
  
  char version[32];
  if (slotStrVal(versionSlot, version, 32)) return errWrongType;

  xmlDocPtr doc = xmlNewDoc(BAD_CAST(version));

  // Assume the native object pointer is the first slot
  SetPtr(objectSlot->uo->slots + 0, doc);

  return errNone;
}

int prXmlDocNewFromFile(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *filenameSlot = g->sp;
  
  char filename[PATH_MAX];
  if (slotStrVal(filenameSlot, filename, PATH_MAX)) return errWrongType;

  xmlDocPtr doc = xmlParseFile(filename);

  // Assume the native object pointer is the first slot
  if (doc != 0)
    SetPtr(objectSlot->uo->slots + 0, doc);
  else
    SetNil(objectSlot);

  return errNone;
}

int prXmlDocFree(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  
  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[0].uptr);

  if (doc == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlDoc wrapper: null pointer\n");
    return errFailed;
  }

  xmlFreeDoc(doc);

  return errNone;
}

int prXmlDocSave(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *filenameSlot = g->sp - 1;
  PyrSlot *formattedSlot = g->sp;
  
  char filename[PATH_MAX];
  if (slotStrVal(filenameSlot, filename, PATH_MAX)) return errWrongType;

  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[0].uptr);
  if (doc == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlDoc wrapper: null pointer\n");
    return errFailed;
  }

  if (!IsTrue(formattedSlot) && !IsFalse(formattedSlot)) return errWrongType;
  bool formatted = IsTrue(formattedSlot);

  FILE *file;

  file = fopen(filename, "w");

  if (file == 0) {
    SetNil(objectSlot);
    return errNone;
  }

  int n = xmlDocFormatDump(file, doc, formatted);

  fclose(file);

  SetInt(objectSlot, n);

  return errNone;
}

int prXmlDocDumpToArray(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  
  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[0].uptr);
  if (doc == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlDoc wrapper: null pointer\n");
    return errFailed;
  }

  xmlChar *memory = 0;
  int size = 0;
  
  xmlDocDumpMemory(doc, &memory, &size);

  // NOT IMPLEMENTED

  xmlFree(memory);

  return errNone;
}

int prXmlDocCreateRootNode(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[0].uptr);
  if (doc == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlDoc wrapper: null pointer\n");
    return errFailed;
  }
  
  char name[1024];
  if (slotStrVal(nameSlot, name, 1024)) return errWrongType;

  xmlNodePtr node = xmlNewDocNode(doc, 0, BAD_CAST(name), 0);

  if (node != 0) {
    xmlDocSetRootElement(doc, node);
    SetPtr(objectSlot, node);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prXmlDocGetRootElement(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  
  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[0].uptr);
  if (doc == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlDoc wrapper: null pointer\n");
    return errFailed;
  }

  xmlNodePtr node = xmlDocGetRootElement(doc);

  if (node != 0)
    SetPtr(objectSlot, node);
  else
    SetNil(objectSlot);

  return errNone;
}

int prXmlNodeGetName(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  
  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  int size = xmlStrlen(node->name);
  PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
  memcpy(strobj->s, node->name, size);
  
  SetObject(objectSlot, strobj);
  g->gc->GCWrite(g->process->interpreter.uo, strobj);

  return errNone;
}

int prXmlNodeGetChildren(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  
  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  if (node->xmlChildrenNode)
    SetPtr(objectSlot, node->xmlChildrenNode);
  else
    SetNil(objectSlot);

  return errNone;
}

int prXmlNodeGetProperty(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *nameSlot = g->sp - 0;

  char name[1024];
  if (slotStrVal(nameSlot, name, 1024)) return errWrongType;

  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  xmlChar *value;
  
  value = xmlGetProp(node, BAD_CAST(name));
  
  if (value != 0) {
    int size = xmlStrlen(BAD_CAST(value));
    PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
    memcpy(strobj->s, BAD_CAST(value), size);
    
    SetObject(objectSlot, strobj);
    g->gc->GCWrite(g->process->interpreter.uo, strobj);
    
    xmlFree(value);
  } else {
    SetNil(objectSlot);
  }
  
  return errNone;
}

int prXmlNodeGetNext(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp;
  
  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  xmlNodePtr next = node->next;
  
  if (next != 0) { 
    SetPtr(objectSlot, next);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prXmlNodeNewChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *nameSlot = g->sp - 1;
  PyrSlot *contentSlot = g->sp;

  char name[1024];
  if (slotStrVal(nameSlot, name, 1024)) return errWrongType;

  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  char *content;

  if (IsNil(contentSlot)) {
    content = 0;
  } else {
    int contentSize = slotStrLen(contentSlot);
    if (contentSize < 0) return errWrongType;
    content = new char[contentSize + 1];

    slotStrVal(contentSlot, content, contentSize + 1);
  }

  xmlNodePtr child = xmlNewChild(node, 0, BAD_CAST(name), BAD_CAST(content));

  if (content) delete [] content;

  if (child != 0) {
    SetPtr(objectSlot, child);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prXmlNodeNewTextChild(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *nameSlot = g->sp - 1;
  PyrSlot *contentSlot = g->sp;

  char name[1024];
  if (slotStrVal(nameSlot, name, 1024)) return errWrongType;

  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  char *content;

  if (IsNil(contentSlot)) {
    content = 0;
  } else {
    int contentSize = slotStrLen(contentSlot);
    if (contentSize < 0) return errWrongType;
    content = new char[contentSize + 1];
    if (slotStrVal(contentSlot, content, contentSize + 1)) {
      delete [] content;
      return errWrongType;
    }
  }

  xmlNodePtr child = xmlNewTextChild(node, 0, BAD_CAST(name), BAD_CAST(content));

  if (content)
    delete [] content;

  if (child != 0) {
    SetPtr(objectSlot, child);
  } else {
    SetNil(objectSlot);
  }

  return errNone;
}

int prXmlNodeSetProperty(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 2;
  PyrSlot *nameSlot = g->sp - 1;
  PyrSlot *valueSlot = g->sp;

  char name[1024];
  if (slotStrVal(nameSlot, name, 1024)) return errWrongType;
  char value[1024];
  if (slotStrVal(valueSlot, value, 1024)) return errWrongType;

  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  if (node == 0) {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }

  xmlSetProp(node, BAD_CAST(name), BAD_CAST(value));

  return errNone;
}

int prXmlNodeGetContent(struct VMGlobals *g, int numArgsPushed) {
  PyrSlot *objectSlot = g->sp - 1;
  PyrSlot *expandSlot = g->sp;

  xmlNodePtr node = (xmlNodePtr)(objectSlot->uo->slots[0].uptr);
  xmlDocPtr doc = (xmlDocPtr)(objectSlot->uo->slots[1].uptr);

  if (node == 0)  {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null pointer\n");
    return errFailed;
  }
  if (doc == 0)  {
    dumpObjectSlot(objectSlot);
    error("invalid xmlNode wrapper: null document pointer\n");
    return errFailed;
  }

  if (!IsTrue(expandSlot) && !IsFalse(expandSlot)) return errWrongType;
  bool expand = IsTrue(expandSlot);

  xmlChar *content;

  content = xmlNodeListGetString(doc, node->xmlChildrenNode, expand);
  
  if (content != 0) {
    int size = xmlStrlen(BAD_CAST(content));
    PyrString *strobj = newPyrStringN(g->gc, size, 0, true);
    memcpy(strobj->s, BAD_CAST(content), size);
    
    SetObject(objectSlot, strobj);
    g->gc->GCWrite(g->process->interpreter.uo, strobj);
    
    xmlFree(content);
  } else {
    SetNil(objectSlot);
  }
  
  return errNone;
}

void initialize_xml2_primitives(void) {
  int base = nextPrimitiveIndex();
  int index = 0;

  definePrimitive(base, index++, "_prXmlDocNew", prXmlDocNew, 2, 0);
  definePrimitive(base, index++, "_prXmlDocNewFromFile", prXmlDocNewFromFile, 2, 0);
  definePrimitive(base, index++, "_prXmlDocFree", prXmlDocFree, 1, 0);
  definePrimitive(base, index++, "_prXmlDocGetRootElement", prXmlDocGetRootElement, 1, 0);
  definePrimitive(base, index++, "_prXmlDocCreateRootNode", prXmlDocCreateRootNode, 2, 0);
  definePrimitive(base, index++, "_prXmlDocSave", prXmlDocSave, 3, 0);
  definePrimitive(base, index++, "_prXmlDocDumpToArray", prXmlDocDumpToArray, 1, 0);

  definePrimitive(base, index++, "_prXmlNodeGetName", prXmlNodeGetName, 1, 0);
  definePrimitive(base, index++, "_prXmlNodeGetProperty", prXmlNodeGetProperty, 2, 0);
  definePrimitive(base, index++, "_prXmlNodeSetProperty", prXmlNodeSetProperty, 3, 0);
  definePrimitive(base, index++, "_prXmlNodeGetChildren", prXmlNodeGetChildren, 1, 0);
  definePrimitive(base, index++, "_prXmlNodeGetNext", prXmlNodeGetNext, 1, 0);
  definePrimitive(base, index++, "_prXmlNodeNewChild", prXmlNodeNewChild, 3, 0);
  definePrimitive(base, index++, "_prXmlNodeNewTextChild", prXmlNodeNewTextChild, 3, 0);
  definePrimitive(base, index++, "_prXmlNodeGetContent", prXmlNodeGetContent, 2, 0);
}
